%{

/* This explicitly includes the platform-specific defines so */
/* that unistd.h can be found.                               */
#include <meta/platform.h>
#include <osd/osd.h>

/* Since the bogomips who wrote flex forgot to put the extern */
/* "C" block around isatty(), even though they go to the      */
/* trouble of redefining it, I will clean up for them.        */
/* Idiots.  Why don't you kick my dog while you're here?      */
extern "C" {
extern int isatty(int);
}

#include <bld/interp.h>
#include <bld/fnop.h>
#include <bld/parse.h>

static char *bldNewString(char *s, const unsigned long l);
static BLD_Operation *bldParseString(char *s, const unsigned long l);
static void bldDebugPrint(char *token, char *text, char *value);
#define	BLD_NEWSTR(x,y)	(bldNewString((char *) (x), strlen((char *) (x)) - (y)))
#define	BLD_STR(s)	yylval.string=(char *)(s)
#define	BLD_SAFESTR(s)	((s) ? (s) : "<null>")

#if DEBUG
#define	BLD_RETURN(token)	bldDebugPrint( #token ,yytext,yylval.string);return token
#define	BLD_ADD(t,s)	WBA_APP->Printf("%s \"%s\"", #t, (char*)s); op->Add(new BLD_##t##Operation(s));
#else
#define	BLD_RETURN(token)	return token
#define	BLD_ADD(t,s)	 op->Add(new BLD_##t##Operation(s));
#endif

%}

%a	6000
%p	5000
%o	8000

scope	[A-Za-z][\_A-Za-z0-9]+
lvar	[A-Za-z][\_A-Za-z0-9]+
rvar	[\_A-Za-z0-9]+
literal [^#$!&|;:<>='"`(){}\[\], \t\n][^#$!&|;:<>='"`(){}\[\], \t\n]*
single	\'[^\'\n]*[^\\\'\n]\'
double	\"[^\"\n]*\"
back	\`[^\`\n]*\`
target	[^\n=#:]+

%x CMT
%x SCP
%x VAR
%x LEFTY
%x RIGHTY
%x SPC

%%

<CMT>.*			{ BLD_RETURN(COMMENT); }
<CMT>\n			{
					BEGIN INITIAL;
					BLD_STR(0);
					++BLD_Interpreter::lineNumber; BLD_RETURN(NEWLINE);
				}

<SCP>{scope}	{
					BEGIN INITIAL;
					BLD_STR(BLD_NEWSTR(yytext, 0));
					BLD_RETURN(NAME);
				}

<VAR>{rvar}				{
							BLD_STR(BLD_NEWSTR(yytext, 0));
							BLD_RETURN(VARIABLE);
						}
<VAR>{scope}\.{rvar}	{
							BLD_STR(BLD_NEWSTR(yytext, 0));
							BLD_RETURN(VARIABLE);
						}
<VAR>[ \t]*=[ \t]*		{ BEGIN INITIAL; BLD_STR(0); BLD_RETURN(EQUAL); }
<VAR>[ \t]*\+=[ \t]*	{ BEGIN INITIAL; BLD_STR(0); BLD_RETURN(PLUS_EQUAL); }
<VAR>[ \t]*\-=[ \t]*	{ BEGIN INITIAL; BLD_STR(0); BLD_RETURN(MINUS_EQUAL); }
<VAR>[ \t]+				{ BLD_STR(0); BLD_RETURN(BLANK_SPACE); }
<VAR>\n					{
							BLD_STR(0);
							++BLD_Interpreter::lineNumber;
							BLD_RETURN(NEWLINE);
						}

\n/[ \t\n]*\{	{ BEGIN LEFTY; ++BLD_Interpreter::lineNumber; }
<LEFTY>[ \t]+	{ ; }
<LEFTY>\n		{ ++BLD_Interpreter::lineNumber; }
<LEFTY>"{"		{ BEGIN INITIAL; BLD_RETURN(LEFT_CURLY); }

\}/[ \t\n]+else					{ BEGIN RIGHTY; BLD_RETURN(RIGHT_CURLY); }
<RIGHTY>[ \t]+					{ ; }
<RIGHTY>\n						{ ++BLD_Interpreter::lineNumber; }
<RIGHTY>else					{ BEGIN INITIAL; BLD_RETURN(ELSE); }
<RIGHTY>else[ \t]*if/[ \t]*\(	{ BEGIN INITIAL; BLD_RETURN(ELSEIF); }

<SPC>":"	{ BEGIN INITIAL; BLD_RETURN(SPEC); }
<SPC>":-"	{ BEGIN INITIAL; BLD_RETURN(PRESPEC); }
<SPC>":x"	{ BEGIN INITIAL; BLD_RETURN(FAILSPEC); }
<SPC>":+"	{ BEGIN INITIAL; BLD_RETURN(POSTSPEC); }
<SPC>[ \t]+	{ ; }
<SPC>\n		{ ++BLD_Interpreter::lineNumber; }

";"			{ BLD_STR(0); BLD_RETURN(SEMI_COLON); }
","			{ BLD_STR(0); BLD_RETURN(COMMA); }
"{"			{ BLD_STR(0); BLD_RETURN(LEFT_CURLY); }
"}"			{ BLD_STR(0); BLD_RETURN(RIGHT_CURLY); }
"("			{ BLD_STR(0); BLD_RETURN(LEFT_PAREN); }
")"			{ BLD_STR(0); BLD_RETURN(RIGHT_PAREN); }
"["			{ BLD_STR(0); BLD_RETURN(LEFT_BRACE); }
"]"			{ BLD_STR(0); BLD_RETURN(RIGHT_BRACE); }
[ \t]*"!"	{ BLD_STR(0); BLD_RETURN(NOT); }
[ \t]*"&&"	{ BLD_STR(0); BLD_RETURN(AND); }
[ \t]*"||"	{ BLD_STR(0); BLD_RETURN(OR); }
[ \t]*"=="	{ BLD_STR(0); BLD_RETURN(EQ); }
[ \t]*"!="	{ BLD_STR(0); BLD_RETURN(NE); }
[ \t]*"<<"	{ BLD_STR(0); BLD_RETURN(LTLT); }
[ \t]*"<"	{ BLD_STR(0); BLD_RETURN(LT); }
[ \t]*"<="	{ BLD_STR(0); BLD_RETURN(LE); }
[ \t]*">>"	{ BLD_STR(0); BLD_RETURN(GTGT); }
[ \t]*">"	{ BLD_STR(0); BLD_RETURN(GT); }
[ \t]*">="	{ BLD_STR(0); BLD_RETURN(GE); }

include		{ BLD_STR(0); BLD_RETURN(INCLUDE); }
build		{ BLD_STR(0); BLD_RETURN(BUILD); }
cd			{ BLD_STR(0); BLD_RETURN(BUILTIN_CD); }
echo		{ BLD_STR(0); BLD_RETURN(BUILTIN_ECHO); }
system		{ BLD_STR(0); BLD_RETURN(SYSTEM); }
check		{ BLD_STR(0); BLD_RETURN(CHECK); }
tcheck		{ BLD_STR(0); BLD_RETURN(TCHECK); }
fcheck		{ BLD_STR(0); BLD_RETURN(FCHECK); }
exists		{ BLD_STR(0); BLD_RETURN(EXISTS); }
dateof		{ BLD_STR(0); BLD_RETURN(DATEOF); }
newest		{ BLD_STR(0); BLD_RETURN(NEWEST); }
latest		{ BLD_STR(0); BLD_RETURN(LATEST); }

begin[ \t]*	{ BEGIN SCP; BLD_RETURN(BEG); }
end	{ BLD_RETURN(END); }

if/[ \t]*\(				{ BLD_RETURN(IF); }
else					{ BLD_RETURN(ELSE); }
else[ \t]*if/[ \t]*\(	{ BLD_RETURN(ELSEIF); }
foreach					{ BLD_RETURN(FOREACH); }

pragma[ \t]+dump			{ BLD_RETURN(PRAGMA_DUMP); }
pragma[ \t]+dump[ \t]+top	{ BLD_RETURN(PRAGMA_DUMPTOP); }
pragma[ \t]+env				{ BLD_RETURN(PRAGMA_ENV); }

set				{ BEGIN VAR; BLD_RETURN(SET); }
break			{ BLD_RETURN(BREAK); }
exit			{ BLD_RETURN(EXIT); }

^[ \t]*{target}/:		{
					BEGIN SPC;
					BLD_STR(BLD_NEWSTR(yytext, 0));
					//fprintf(stderr, "TARGET %s\n", yylval.string);
					BLD_RETURN(TARGET);
				}
^[ \t]*{target}/:\-	{
					BEGIN SPC;
					BLD_STR(BLD_NEWSTR(yytext, 0));
					//fprintf(stderr, "TARGET %s\n", yylval.string);
					BLD_RETURN(TARGET);
				}
^[ \t]*{target}/:\+	{
					BEGIN SPC;
					BLD_STR(BLD_NEWSTR(yytext, 0));
					//fprintf(stderr, "TARGET %s\n", yylval.string);
					BLD_RETURN(TARGET);
				}

\$\({rvar}\)			{
							BLD_STR(BLD_NEWSTR(yytext + 2, 1));
							BLD_RETURN(VARIABLE);
						}
\$\({scope}\.{rvar}\)	{
							BLD_STR(BLD_NEWSTR(yytext + 2, 1));
							BLD_RETURN(VARIABLE);
						}
{literal}				{
							BLD_STR(BLD_NEWSTR(yytext, 0));
							//fprintf(stderr, "LITERAL %s\n", yylval.string);
							BLD_RETURN(LITERAL);
						}

\'\'						{
								yylval.op = new BLD_LiteralOperation("");
								BLD_RETURN(STRING);
							}
\'[^\'\n]\'					{
								yylval.op = new BLD_LiteralOperation
													(*(yytext + 1));
								BLD_RETURN(STRING);
							}
\'([^\\\']*\\.)*[^\']*\'	{
								yylval.op = bldParseString((char *)yytext,
												strlen((const char *)yytext));
								BLD_RETURN(STRING);
							}

\"\"						{
								yylval.op = new BLD_LiteralOperation("");
								BLD_RETURN(STRING);
							}
\"[^\"\n]\"					{
								yylval.op = new BLD_LiteralOperation
													(*(yytext + 1));
								BLD_RETURN(STRING);
							}
\"([^\\\"]*\\.)*[^\"]*\"	{
								yylval.op = bldParseString((char *)yytext,
												strlen((const char *)yytext));
								BLD_RETURN(STRING);
							}

\`\`						{
								yylval.op = new BLD_LiteralOperation("");
								BLD_RETURN(STRING);
							}
\`[^\`\n]\`					{
								OpArray ops(1);
								ops.AppendElement(new BLD_LiteralOperation
														(*(yytext + 1)));
								yylval.op = new BLD_SystemOperation(ops, TRUE);
								BLD_RETURN(STRING);
							}
\`([^\\\`]*\\.)*[^\`]*\`	{
								yylval.op = bldParseString((char *)yytext,
												strlen((const char *)yytext));
								BLD_RETURN(STRING);
							}

"#"		{ BEGIN CMT; }
\\\n+	{ ++BLD_Interpreter::lineNumber; }
\n		{ BLD_STR(0); ++BLD_Interpreter::lineNumber; BLD_RETURN(NEWLINE); }
[\t ]+	{ BLD_STR(0); BLD_RETURN(BLANK_SPACE); }

%%

int yywrap(void)
{
	return 1;
}

void yyerror(char *errmsg)
{
	fprintf(stderr,"%s on \"%s\" in %s at line %d\n", errmsg,
			(yytext ? (char *) yytext : "null"),
			BLD_Interpreter::parseFileName, BLD_Interpreter::lineNumber);
	BLD_Interpreter::parseSuccess = FALSE;
}

static char *bldNewString(char *s, const unsigned long l)
{
	register char *cp = new char[l + 1];
	osdMemcpy(cp, s, l);
	cp[l] = 0;
	return cp;
}

static BLD_Operation *bldParseString(char *s, const unsigned long l)
{
	BLD_ListOperation *listOp = new BLD_ListOperation();
	register BLD_ListOperation *op = listOp;
	GPL_Boolean isSingleTick = FALSE;
	GPL_Boolean isDoubleTick = FALSE;
	GPL_Boolean isBackTick = FALSE;
	GPL_Boolean isEscaped = FALSE;
	GPL_String string;

	for (register unsigned long i = 0; l > i; ++i)
	{
		if (isEscaped)
		{
			string.Append(s[i]);
			isEscaped = FALSE;
			continue;
		}

		switch (s[i])
		{
		case '\\':
			if (!isSingleTick || (l > i + 1 && '\'' == s[i + 1]))
			{
				isEscaped = TRUE;
			}
			else
			{
				string.Append(s[i]);
			}
			break;

		case '\'':
			if (!isSingleTick)
			{
				if (isDoubleTick || isBackTick)
				{
					if (0 < string.GetLength())
					{
						BLD_ADD(String, string);
					}
					string.Clear();
				}
				isSingleTick = TRUE;
			}
			else
			{
				if (0 < string.GetLength())
				{
					BLD_ADD(Literal, string);
				}
				string.Clear();
				isSingleTick = FALSE;
			}
			break;

		case '"':
			if (isSingleTick)
			{
				string.Append(s[i]);
			}
			else if (!isDoubleTick)
			{
				if (0 < string.GetLength())
				{
					if (isBackTick)
					{
						BLD_ADD(String, string);
					}
					else
					{
						BLD_ADD(Literal, string);
					}
				}
				string.Clear();
				isDoubleTick = TRUE;
			}
			else
			{
				if (0 < string.GetLength())
				{
					BLD_ADD(String, string);
				}
				string.Clear();
				isDoubleTick = FALSE;
			}
			break;

		case '`':
			if (isSingleTick)
			{
				string.Append(s[i]);
			}
			else if (!isBackTick)
			{
				if (0 < string.GetLength())
				{
					if (isDoubleTick)
					{
						BLD_ADD(String, string);
					}
					else
					{
						BLD_ADD(Literal, string);
					}
				}
				string.Clear();
				op = new BLD_ListOperation();
				isBackTick = TRUE;
			}
			else
			{
				if (0 < string.GetLength())
				{
					BLD_ADD(String, string);
				}
				string.Clear();
				OpArray ops(1);
				ops.AppendElement(op);
				listOp->Add(new BLD_SystemOperation(ops, TRUE));
				op = listOp;
				isBackTick = FALSE;
			}
			break;

		default:
			string.Append(s[i]);
		}
	}

	if (0 < string.GetLength())
	{
		BLD_ADD(Literal, string);
	}

	return (BLD_Operation *) listOp;
}

static void bldDebugPrint(char *token, char *text, char *value)
{
	if (!value)
	{
		printf("%s: \"%s\"\n", BLD_SAFESTR(token), BLD_SAFESTR(text));
	}
	else
	{
		printf("%s: \"%s\"=>\"%s\"\n", BLD_SAFESTR(token), BLD_SAFESTR(text), BLD_SAFESTR(value));
	}
	fflush(stdout);
}

#if OSD_OSTYPE==OSD_WIN32
int isatty(int fildes)
{
	return !osdFileIsDiskFile((OSD_FILE_HANDLE) fildes);
}
#endif /* OSD_OSTYPE==OSD_WIN32 */
