#!/usr/local/bin/perl -w

require 5.000;
use strict;
use English;
use FileHandle;
require Getopt::Long;

my($debug) = 0;
my($inline) = 1;
my(@elements) = ();

sub processInlines
{
	my($line) = shift;
	$line =~ s/\\\&\././g;
	$line =~ s/\\fB(.*?)((\\f[A-Z])|$)/bf($::1)/g;
	$line =~ s/\\fI(.*?)((\\f[A-Z])|$)/em($::1)/g;
	return $line;
}

sub readFile()
{
	my($fileName) = shift;
	my($elements) = shift;

	if (open(INPUT, "<$fileName"))
	{
		my($outputFileName) = $fileName;
		$outputFileName =~ s/\.man\s*$/.yo/;

		my($inComment) = $inline;
		my($inSynopsis) = 0;
		my($inSystemCalls) = 0;
		my($inEnum) = 0;
		my($name);
		my($text);
		my($string);
		my($line);
		while (defined($line = <INPUT>))
		{
			# Strip leading and trailing white-space.
			$line =~ s/^\s*//;
			$line =~ s/\s*$//;

			if ($inComment)
			{
				if ($line =~ m/^BEGIN_INLINE_MANPAGE/)
				{
					$inComment = 0;
				}
			}
			else
			{
				if ($line =~ m/^END_INLINE_MANPAGE/)
				{
#STDOUT->print("\$outputFileName == \'$outputFileName\' => $#{$elements}\n");
					&writeFile($outputFileName, $elements);
					@{$elements} = ();
					$inComment = 1;
					next;
				}

				# Pre-process any embedded fonts.
				$line = &processInlines($line);

#STDOUT->print("\$line == \'$line\'\n");

				study($line);

				if ($line =~ m/^\.TH\s+\"([^"]*)\"\s+\"([^"]*)\"\s+\"([^"]*)\"\s+\"([^"]*)\"\s+\"([^"]*)\"/)
				{
					$text = "manpage($::1)($::2)($::3)($::4)($::5)";
				}
				elsif ($line =~ m/^\.RS/)
				{
					$inEnum = 1;
					push(@{$elements}, 'startenum()');
					$string = 'enum()(';
				}
				elsif ($line =~ m/^\.RE/)
				{
					$inEnum = 0;
					if (0 < length($string))
					{
						push(@{$elements}, "$string\)");
					}
					$text = 'endenum()';
				}
				elsif ($line =~ m/^\.SH\s+"?NAME"?/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					while (defined($line = <INPUT>) && $line =~ m/^\s*$/)
					{
					}
					$line =~ s/\\f[A-Z]//g;
#STDOUT->print("\$line == \'$line\'\n");
					if ($line =~ m/([^\-\s]+)\s*\\?\-\s*(.*)/)
					{
						$outputFileName = $::1;
						$outputFileName .= '.yo';
#STDOUT->print("\$outputFileName == \'$outputFileName\'\n");
						$text = "manpagename($::1)($::2)";
					}
					else
					{
						$text = 'manpagename()';
					}
				}
				elsif ($line =~ m/^\.SH\s+"?CLASS"?/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					while (defined($line = <INPUT>) && $line =~ m/^\s*$/)
					{
					}
					$line =~ s/\\f[A-Z]//g;
#STDOUT->print("\$line == \'$line\'\n");
					if ($line =~ m/^(\.[A-Z]+)?\s*([^\-\s]+)/)
					{
						$name = $::2;
STDOUT->print("\$name == \'$name\'\n");
						$outputFileName = $name;
						$outputFileName .= '.yo';
						if ($line =~ m/\\?\-\s*(.+)$/ or
							(defined($line = <INPUT>) && $line =~ m/^\\?\-\s*(.+)/))
						{
							$text = "manpageclass($name)($::1)";
						}
						else
						{
							$text = "manpageclass($name)";
						}
					}
					else
					{
						$text = 'manpageclass()';
					}
				}
				elsif ($line =~ m/^\.SH\s+"?SYNOPSIS"?\s*$/)
				{
					$inSynopsis = 1;
					$inSystemCalls = 0;
				}
				elsif ($line =~ m/^\.SH\s+"?DESCRIPTION"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'manpagesynopsis()';
				}
				elsif ($line =~ m/^\.SH\s+"?FILE"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'manpagefiles()';
				}
				elsif ($line =~ m/^\.SH\s+"?SEE\s+ALSO"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'manpageseealso()';
				}
				elsif ($line =~ m/^\.SH\s+"?(WARNINGS|DIAGNOSTICS)"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'manpagediagnostics()';
				}
				elsif ($line =~ m/^\.SH\s+"?BUGS"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'manpagebugs()';
				}
				elsif ($line =~ m/^\.SH\s+"?AUTHOR"?\s*$/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = 'endmanpage()
				elsif ($line =~ m/^\.SH\s+.*SYSTEM\s*CALL/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 1;
				}
				elsif ($line =~ m/^\.SH\s+"?([^"]*)"?/)
				{
					$inSynopsis = 0;
					$inSystemCalls = 0;
					$text = "manpagesection($::1)";
				}
				elsif ($inSynopsis)
				{
					next;
				}
				elsif ($inSystemCalls)
				{
					next;
				}
				elsif ($line =~ m/^\.[IP]P(?:(?:\s+(.*))|$)/)
				{
					$text = "$::1";
				}
				elsif ($line =~ m/^\.br/)
				{
					$text = 'nl()';
				}
				elsif ($line =~ m/^\.B\s+(.*)/)
				{
					$text = "bf($::1)";
				}
				elsif ($line =~ m/^\.I\s+(.*)/)
				{
					$text = "em($::1)";
				}
				elsif ($line =~ m/^\.BR\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+(.*))?)?)?)?)?/)
				{
					if (defined($::6))
					{
						$text = "bf($::1) $::2 bf($::3) $::4 bf($::5) $::6";
					}
					elsif (defined($::5))
					{
						$text = "bf($::1) $::2 bf($::3) $::4 bf($::5)";
					}
					elsif (defined($::4))
					{
						$text = "bf($::1) $::2 bf($::3) $::4";
					}
					elsif (defined($::3))
					{
						$text = "bf($::1) $::2 bf($::3)";
					}
					elsif (defined($::2))
					{
						$text = "bf($::1) $::2";
					}
					elsif (defined($::1))
					{
						$text = "bf($::1)";
					}
				}
				elsif ($line =~ m/^\.IR\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+((?:"[^"]*")|(?:\S+))(?:\s+(.*))?)?)?)?)?/)
				{
					if (defined($::6))
					{
						$text = "em($::1) $::2 em($::3) $::4 em($::5) $::6";
					}
					elsif (defined($::5))
					{
						$text = "em($::1) $::2 em($::3) $::4 em($::5)";
					}
					elsif (defined($::4))
					{
						$text = "em($::1) $::2 em($::3) $::4";
					}
					elsif (defined($::3))
					{
						$text = "em($::1) $::2 em($::3)";
					}
					elsif (defined($::2))
					{
						$text = "em($::1) $::2";
					}
					elsif (defined($::1))
					{
						$text = "em($::1)";
					}
				}
				elsif ($line =~ m/^\s*$/)
				{
					if ($inEnum)
					{
						if (0 < length($string))
						{
							push(@{$elements}, "$string\)");
						}
						$string = 'enum()(';
					}
					else
					{
						$text = $line;
					}
				}
				else
				{
					$text = $line;
				}
				if ($inEnum)
				{
					$string .= "\n" . $text;
				}
				else
				{
					push(@{$elements}, $text);
				}
			}
		}
		close(INPUT);
		unless (-1 == $#{$elements})
		{
			&writeFile($outputFileName, $elements);
			@{$elements} = ();
		}
	}
	else
	{
		STDERR->print("Unable to open $fileName\n");
	}
}

sub writeFile()
{
	my($fileName) = shift;
	my($elements) = shift;
#STDOUT->print("\$#{\$elements} == $#{$elements}\n");

	if (open(OUTPUT, ">$fileName"))
	{
		OUTPUT->print("includefile(../meta/manfmt.yo)\n");
		my($element);
		my($hasAuthor) = 0;
		foreach $element (@{$elements})
		{
			if ($element =~ m/manpageauthor\(\)/)
			{
				$hasAuthor = 1;
			}
			STDOUT->print("$element\n") if $debug;
			OUTPUT->print("$element\n");
		}
		unless ($hasAuthor)
		{
			OUTPUT->print("manpageauthor()\n");
			OUTPUT->print(('aj' eq $ENV{'LOGNAME'}) ? "AJ Weber" : "man2yodl.pl");
			OUTPUT->print("\n");
		}
		close(OUTPUT);
		STDOUT->print("Wrote $fileName\n");
	}
	else
	{
		STDERR->print("Unable to open $fileName\n");
	}
}

# Parse the command-line options.
$Getopt::Long::autoabbrev = 1 unless $Getopt::Long::autoabbrev;
&Getopt::Long::GetOptions('debug!' => \$debug, 'inline!' => \$inline);
unless (0 <= $#::ARGV)
{
	STDERR->print("Usage: $::PROGRAM_NAME [-debug][-inline] <file>...\n");
	exit(-1);
}

# Read the files.
while (0 <= $#::ARGV)
{
	&readFile($::ARGV[0], \@elements);
	shift(@::ARGV);
}

exit(0);

__END__
